{
module Parser where

import Data.Char
}
%name parse
%tokentype { Token }
%error { parseError }
%token
	var	{ TokenVar $$ }
	int	{ TokenInt $$ }
	'+'	{ TokenPlus }
	'-'	{ TokenMinus }
	eq	{ TokenEQ }
	neq	{ TokenNEQ }
	lt	{ TokenLT }
	lte	{ TokenLTE }
	gt	{ TokenGT }
	gte	{ TokenGTE }
	'('	{ TokenOB }
	')'	{ TokenCB }
	','	{ TokenComma }
	';'	{ TokenSemi }
    upd { TokenUpd }
    sel { TokenSel }
    map { TokenMap }
    bot { TokenBot }
    mapsto { TokenMapsTo }
%%
All : Map Terms            { ($1, $2) }

Map : map '(' MapTerms ')' ';'  { Map $3 }

MapTerms 
    : MapTerm ',' MapTerms      { $1 : $3 }
    | MapTerm                   { [$1] }

MapTerm : Atom mapsto Expr      { ($1, $3) }

Terms 	
    : Term ';' Terms            { $1 : $3 }
	| Term ';'	                { [$1] }

Term
    : Expr                      { Expr $1 }
    | Pred                      { Pred $1 }

Pred
    : Expr lt Expr              { Lt $1 $3 }
    | Expr gt Expr              { Gt $1 $3 }
    | Expr gte Expr             { Gte $1 $3 }
    | Expr lte Expr             { Lte $1 $3 }
    | Expr eq Expr              { Eq $1 $3 }
    | Expr neq Expr             { Neq $1 $3 }

Expr	
	: Expr '+' Expr             { Plus $1 $3 }
	| Expr '-' Expr	            { Minus $1 $3 }
    | upd '(' Expr ',' Expr ',' Expr ')'   { Upd $3 $5 $7 }
    | sel '(' Expr ',' Expr ')' { Sel $3 $5 }
    | FnCall                    { $1 }
	| Atom	                    { Atom $1 }

FnCall
    : var '(' FnArgs ')'        { Call (Var $1) $3 }

FnArgs
    : Expr ',' FnArgs          { $1 : $3 }
    | Expr                     { [$1] }

Atom
	: int 	                    { Int $1 }
    | '-' int                   { Int (-$2) }
	| var	                    { Var $1 }
    | bot                       { Bot }
{
parseError :: [Token] -> a
parseError _ = error "Parse error"

data Map = Map [(Atom, Expr)]

data Term
    = Expr Expr
    | Pred Pred
   deriving (Eq,Show)

data Pred
    = Lt Expr Expr
    | Lte Expr Expr
    | Gt Expr Expr
    | Gte Expr Expr
    | Eq Expr Expr
    | Neq Expr Expr
   deriving (Eq,Show)

data Expr
	= Plus Expr Expr
	| Minus Expr Expr
    | Upd Expr Expr Expr
    | Sel Expr Expr
	| Atom Atom
    | Call Atom [Expr]
   deriving (Eq,Show)

data Atom
	= Int Int
	| Var String
    | Bot
   deriving (Eq,Show)

data Token
	= TokenInt Int
	| TokenVar String
	| TokenPlus | TokenMinus
    | TokenGT | TokenLT | TokenGTE | TokenLTE | TokenEQ | TokenNEQ
    | TokenUpd | TokenSel
	| TokenOB | TokenCB 
    | TokenMap
	| TokenSemi | TokenComma | TokenMapsTo | TokenBot
   deriving Show

lexer :: String -> [Token]
lexer [] = []
lexer (c:cs) 
      | isSpace c = lexer cs
      | isVarStart c = lexVar (c:cs)
      | isDigit c = lexNum (c:cs)
lexer ('%':cs) = lexCom cs
lexer ('\n':cs) = lexer cs
lexer ('#':cs) = TokenBot : lexer cs
lexer (';':cs) = TokenSemi : lexer cs
lexer (',':cs) = TokenComma : lexer cs
lexer ('+':cs) = TokenPlus : lexer cs
lexer ('-':'>':cs) = TokenMapsTo : lexer cs
lexer ('-':cs) = TokenMinus : lexer cs
lexer ('!':('=':cs)) = TokenNEQ : lexer cs
lexer ('=':cs) = TokenEQ : lexer cs
lexer ('<':('=':cs)) = TokenLTE : lexer cs
lexer ('>':('=':cs)) = TokenGTE : lexer cs
lexer ('<':cs) = TokenLT : lexer cs
lexer ('>':cs) = TokenGT : lexer cs
lexer ('(':cs) = TokenOB : lexer cs
lexer (')':cs) = TokenCB : lexer cs
lexer fail = error ("Lexer failed on: " ++ fail)

isVarStart c = isAlpha c || c == '_'
isVarChar c = isAlpha c || isDigit c || c == '\'' || c == '_'

lexCom cs = lexer rest
      where (_,rest) = span (\x -> not (x == '\n')) cs

lexNum cs = TokenInt (read num) : lexer rest
      where (num,rest) = span isDigit cs

lexVar cs =
   case span isVarChar cs of
      ("upd",rest) -> TokenUpd : lexer rest
      ("sel",rest) -> TokenSel : lexer rest
      ("map",rest) -> TokenMap : lexer rest
      (var,rest)   -> TokenVar var : lexer rest

}
